# -*- coding: utf-8 -*-
import typing

from apispec.utils import OpenAPIVersion
import serpyco

import dataclasses

__location_map__ = {
    "query": "query",
    "querystring": "query",
    "json": "body",
    "headers": "header",
    "cookies": "cookie",
    "form": "formData",
    "files": "formData",
}


class OpenAPIConverter(object):
    """Converter generating OpenAPI specification from serpyco schemas and fields

    :param str|OpenAPIVersion openapi_version: The OpenAPI version to use.
        Should be in the form '2.x' or '3.x.x' to comply with the OpenAPI standard.
    """

    def __init__(self, openapi_version):
        self.openapi_version = OpenAPIVersion(openapi_version)
        # Schema references
        self.refs = {}

    def get_ref_path(self):
        """Return the path for references based on the openapi version"""
        ref_paths = {2: "definitions", 3: "components/schemas"}
        return ref_paths[self.openapi_version.major]

    def schema2jsonschema(self, schema, **kwargs):
        return self.fields2jsonschema(dataclasses.fields(schema), schema, **kwargs)

    def resolve_schema_dict(self, schema):
        if isinstance(schema, dict):
            if schema.get("type") == "array" and "items" in schema:
                schema["items"] = self.resolve_schema_dict(schema["items"])
            if schema.get("type") == "object" and "properties" in schema:
                schema["properties"] = {
                    k: self.resolve_schema_dict(v)
                    for k, v in schema["properties"].items()
                }
            return schema

        if schema in self.refs:
            ref_path = self.get_ref_path()
            ref_schema = {"$ref": "#/{0}/{1}".format(ref_path, self.refs[schema])}
            return ref_schema

        return self.schema2jsonschema(schema)

    def fields2jsonschema(self, fields, schema=None):
        """Convert dataclass field into json_schema"""
        field_names = [field.name for field in fields]
        serializer = serpyco.Serializer(dataclass=schema, only=field_names)

        return serializer.json_schema()

    def schema2parameters(self, schema, **kwargs):
        """Return an array of OpenAPI parameters given a given dataclass.
        If `default_in` is "body", then return an array
        of a single parameter; else return an array of a parameter for each included field in
        the dataclass.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject
        """
        return self.fields2parameters(dataclasses.fields(schema), schema, **kwargs)

    def fields2parameters(
        self,
        fields,
        schema,
        default_in="body",
        name="body",
        required=False,
        description=None,
        **kwargs
    ):
        """Convert dataclass fields into OpenAPI parameters"""
        openapi_default_in = __location_map__.get(default_in, default_in)
        if self.openapi_version.major < 3 and openapi_default_in == "body":
            prop = self.resolve_schema_dict(schema)

            param = {
                "in": openapi_default_in,
                "required": required,
                "name": name,
                "schema": prop,
            }

            if description:
                param["description"] = description

            return [param]

        parameters = []
        body_param = None
        for field in fields:

            field_name = field.name
            field_obj = field

            param = self.field2parameter(
                schema, field_obj, name=field_name, default_in=default_in
            )
            if (
                self.openapi_version.major < 3
                and param["in"] == "body"
                and body_param is not None
            ):
                body_param["schema"]["properties"].update(param["schema"]["properties"])
                required_fields = param["schema"].get("required", [])
                if required_fields:
                    body_param["schema"].setdefault("required", []).extend(
                        required_fields
                    )
            else:
                if self.openapi_version.major < 3 and param["in"] == "body":
                    body_param = param
                parameters.append(param)
        return parameters

    def field2parameter(self, schema, field, name="body", default_in="body"):
        """Return an OpenAPI parameter as a `dict`, given a dataclass field.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject
        """
        assert schema

        serializer = serpyco.Serializer(schema, only=[field.name])
        field_json_schema = serializer.json_schema()["properties"][field.name]

        return self.property2parameter(
            field_json_schema,
            name=name,
            required=isinstance(field.default, dataclasses._MISSING_TYPE),
            multiple=isinstance(field.type, typing.Sequence),
            default_in=default_in,
        )

    def property2parameter(
        self,
        prop,
        name="body",
        required=False,
        multiple=False,
        location=None,
        default_in="body",
    ):
        """Return the Parameter Object definition for a JSON Schema property.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject

        :param dict prop: JSON Schema property
        :param str name: Field name
        :param bool required: Parameter is required
        :param bool multiple: Parameter is repeated
        :param str location: Location to look for ``name``
        :param str default_in: Default location to look for ``name``
        :raise: TranslationError if arg object cannot be translated to a Parameter Object schema.
        :rtype: dict, a Parameter Object
        """
        openapi_default_in = __location_map__.get(default_in, default_in)
        openapi_location = __location_map__.get(location, openapi_default_in)
        ret = {"in": openapi_location, "name": name}

        if openapi_location == "body":
            ret["required"] = False
            ret["name"] = "body"
            ret["schema"] = {
                "type": "object",
                "properties": {name: prop} if name else {},
            }
            if name and required:
                ret["schema"]["required"] = [name]
        else:
            ret["required"] = required
            if self.openapi_version.major < 3:
                if multiple:
                    ret["collectionFormat"] = "multi"
                ret.update(prop)
            else:
                if multiple:
                    ret["explode"] = True
                    ret["style"] = "form"
                if prop.get("description", None):
                    ret["description"] = prop.pop("description")
                ret["schema"] = prop
        return ret



# -*- coding: utf-8 -*-
"""Serpyco plugin for apispec. Allows passing a python dataclass
to `APISpec.definition <apispec.APISpec.definition>`
and `APISpec.path <apispec.APISpec.path>` (for responses). Note serpyco field type is supported.
"""
import serpyco
from apispec import BasePlugin

from apispec_serpyco.openapi import OpenAPIConverter
from apispec_serpyco.utils import schema_name_resolver


def extract_definitions_from_json_schema(definition):
    definitions = {}

    for name, definition_ in definition.get("definitions", {}).items():

        # TODO BS: Bypass a serpyco bug
        if definition_ is None:
            continue

        definitions[name] = definition_
        if definition_.get("definitions"):
            definitions.update(extract_definitions_from_json_schema(definition_))

    return definitions


def replace_refs_for_openapi3(data):
    for key, value in data.items():
        if isinstance(value, dict):
            replace_refs_for_openapi3(value)
        elif key == "$ref" and value.startswith("#/definitions"):
            data[key] = value.replace("#/definitions", "#/components/schemas")


def replace_auto_refs(schema_name, data, openapi_version):
    for key, value in data.items():
        if isinstance(value, dict):
            replace_auto_refs(schema_name, value, openapi_version)
        elif key == "$ref" and value == "#":
            if openapi_version.major < 3:
                data[key] = "#/definitions/{}".format(schema_name)
            else:
                data[key] = "#/components/schemas/{}".format(schema_name)


class SerpycoPlugin(BasePlugin):
    """APISpec plugin handling python dataclass (with serpyco typing support)"""

    def __init__(self, schema_name_resolver=None):
        super(SerpycoPlugin, self).__init__()
        self.spec = None
        # self.schema_name_resolver = schema_name_resolver
        self.openapi_version = None
        self.openapi = None
        self.schema_name_resolver = schema_name_resolver

    def init_spec(self, spec):
        """Initialize plugin with APISpec object

        :param APISpec spec: APISpec object this plugin instance is attached to
        """
        super(SerpycoPlugin, self).init_spec(spec)
        self.spec = spec
        self.openapi_version = spec.openapi_version
        self.openapi = OpenAPIConverter(openapi_version=spec.openapi_version)

    def schema_helper(self, name, schema=None, **kwargs):
        """Definition helper that allows using a dataclass to provide
        OpenAPI metadata.

        :param type|type schema a dataclass class
        """
        with_definition = kwargs.get("with_definition")

        if schema is None and not with_definition:
            return None

        if with_definition:
            return with_definition

        # Store registered refs, keyed by Schema class
        self.openapi.refs[schema] = name

        builder = serpyco.SchemaBuilder(
            schema, get_definition_name=schema_name_resolver
        )
        json_schema = builder.json_schema()

        if self.openapi_version.major > 2:
            replace_refs_for_openapi3(json_schema["properties"])

        # Replace auto reference to absolute reference
        replace_auto_refs(name, json_schema["properties"], self.openapi_version)

        # If definitions in json_schema, add them
        if json_schema.get("definitions"):
            # FIXME BS 2018-10-16: There is a problematic here: the serializer produce ref with an
            # automatic naming. This naming is used here but with apispec usage, name can be
            # different because specified in name parameter.
            flat_definitions = extract_definitions_from_json_schema(json_schema)
            for name, definition in flat_definitions.items():
                self.spec.components.schema(name, with_definition=definition)

        # Clean json_schema (to be OpenAPI compatible)
        json_schema.pop("definitions", None)
        json_schema.pop("$schema", None)

        return json_schema

    def parameter_helper(self, **kwargs):
        """Parameter component helper that allows using a dataclass
        in parameter definition.

        :param type|type schema: A dataclass.
        """
        # In OpenAPIv3, this only works when using the complex form using "content"
        self.resolve_schema(kwargs)
        return kwargs

    def response_helper(self, **kwargs):
        """Response component helper that allows using a dataclass in response definition.

        :param type|Schema schema: A marshmallow Schema class or instance.
        """
        self.resolve_schema(kwargs)
        return kwargs

    def operation_helper(self, path=None, operations=None, **kwargs):
        """May mutate operations.

        :param str path: Path to the resource
        :param dict operations: A `dict` mapping HTTP methods to operation object.
        """
        for operation in operations.values():
            if not isinstance(operation, dict):
                continue
            if "parameters" in operation:
                operation["parameters"] = self.resolve_parameters(
                    operation["parameters"]
                )
            if self.openapi_version.major >= 3:
                if "requestBody" in operation:
                    self.resolve_schema_in_request_body(operation["requestBody"])
            for response in operation.get("responses", {}).values():
                self.resolve_schema(response)

    def resolve_schema_in_request_body(self, request_body):
        """Function to resolve a schema in a requestBody object - modifies then
        response dict to convert dataclass into dict
        """
        content = request_body["content"]
        for content_type in content:
            schema = content[content_type]["schema"]
            content[content_type]["schema"] = self.openapi.resolve_schema_dict(schema)

    def resolve_schema(self, data):
        """Function to resolve a schema in a parameter or response - modifies the
        corresponding dict to convert dataclass class into dict

        :param APISpec spec: `APISpec` containing refs.
        :param dict data: the parameter or response dictionary that may contain a dataclass
        :param bool dump: Introspect dump logic.
        :param bool load: Introspect load logic.
        """
        if self.openapi_version.major < 3:
            if "schema" in data:
                data["schema"] = self.openapi.resolve_schema_dict(data["schema"])
        else:
            if "content" in data:
                for content_type in data["content"]:
                    schema = data["content"][content_type]["schema"]
                    data["content"][content_type][
                        "schema"
                    ] = self.openapi.resolve_schema_dict(schema)

    def resolve_parameters(self, parameters):
        resolved = []
        for parameter in parameters:
            if not isinstance(parameter.get("schema", {}), dict):
                schema = parameter["schema"]
                if "in" in parameter:
                    del parameter["schema"]
                    resolved += self.openapi.schema2parameters(
                        schema, default_in=parameter.pop("in"), **parameter
                    )
                    continue
            self.resolve_schema(parameter)
            resolved.append(parameter)
        return resolved



# coding: utf-8
import dataclasses


def schema_name_resolver(dataclass_, only=None, exclude=None) -> str:
    # TODO BS 2018-11-27: Prevent Serpyco bug (who can call here with scalar types)
    if not dataclasses.is_dataclass(dataclass_):
        return dataclass_.__name__

    dataclass_name = dataclass_.__name__
    only = only or []
    exclude = exclude or []
    excluded_field_names = exclude
    dataclass_field_names = [f.name for f in dataclasses.fields(dataclass_)]

    for dataclass_field_name in dataclass_field_names:
        if dataclass_field_name not in only:
            excluded_field_names.append(dataclass_field_name)

    if not excluded_field_names:
        return dataclass_name

    return "{}_exclude_{}".format(dataclass_name, "_".join(excluded_field_names))



# coding: utf-8
from setuptools import setup

setup(
    name="apispec_serpyco",
    version="0.4",
    author="Algoo",
    author_email="contact@algoo.fr",
    description="Serpyco plugin for Apispec",
    license="MIT",
    keywords="apispec openapi serpyco api",
    url="https://github.com/algoo/apispec_plugins",
    packages=["apispec_serpyco"],
    long_description="https://github.com/algoo/apispec_plugins/tree/master/apispec_serpyco",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=["apispec>=1.0.0b1", "serpyco"],
    extras_require={"test": ["pytest"]},
)
